from __future__ import absolute_import
import logging

from opentracing.ext import tags
from tornado.stack_context import wrap as keep_stack_context

from opentracing_instrumentation import utils
from ..request_context import get_current_span, span_in_stack_context
from ._patcher import Patcher


try:
    from boto3.resources.action import ServiceAction
    from boto3.s3 import inject as s3_functions
    from botocore import xform_name
    from botocore.client import BaseClient
    from botocore.exceptions import ClientError
    from s3transfer.futures import BoundedExecutor
except ImportError:
    pass
else:
    _service_action_call = ServiceAction.__call__
    _client_make_api_call = BaseClient._make_api_call
    _Executor = BoundedExecutor.EXECUTOR_CLS

logger = logging.getLogger(__name__)


class Boto3Patcher(Patcher):
    applicable = '_service_action_call' in globals()

    S3_FUNCTIONS_TO_INSTRUMENT = (
        'copy',
        'download_file',
        'download_fileobj',
        'upload_file',
        'upload_fileobj',
    )

    def __init__(self):
        super(Boto3Patcher, self).__init__()
        self.s3_original_funcs = {}

    def _install_patches(self):
        ServiceAction.__call__ = self._get_service_action_call_wrapper()
        BaseClient._make_api_call = self._get_client_make_api_call_wrapper()
        BoundedExecutor.EXECUTOR_CLS = self._get_instrumented_executor_cls()
        for func_name in self.S3_FUNCTIONS_TO_INSTRUMENT:
            func = getattr(s3_functions, func_name, None)
            if func:
                self.s3_original_funcs[func_name] = func
                func_wrapper = self._get_s3_call_wrapper(func)
                setattr(s3_functions, func_name, func_wrapper)
            else:
                logging.warning('S3 function %s not found', func_name)

    def _reset_patches(self):
        ServiceAction.__call__ = _service_action_call
        BaseClient._make_api_call = _client_make_api_call
        BoundedExecutor.EXECUTOR_CLS = _Executor
        for func_name, original_func in self.s3_original_funcs.items():
            setattr(s3_functions, func_name, original_func)

    @staticmethod
    def set_request_id_tag(span, response):
        metadata = response.get('ResponseMetadata')

        # there is no ResponseMetadata for
        # boto3:dynamodb:describe_table
        if metadata:
            request_id = metadata.get('RequestId')

            # when using boto3.client('s3')
            # instead of boto3.resource('s3'),
            # there is no RequestId for
            # boto3:s3:CreateBucket
            if request_id:
                span.set_tag('aws.request_id', request_id)

    def _get_service_action_call_wrapper(self):
        def service_action_call_wrapper(service, parent, *args, **kwargs):
            """Wraps ServiceAction.__call__"""

            service_name = parent.meta.service_name
            operation_name = xform_name(
                service._action_model.request.operation
            )

            return self.perform_call(
                _service_action_call, 'resource',
                service_name, operation_name,
                service, parent, *args, **kwargs
            )

        return service_action_call_wrapper

    def _get_client_make_api_call_wrapper(self):
        def make_api_call_wrapper(client, operation_name, api_params):
            """Wraps BaseClient._make_api_call"""

            service_name = client._service_model.service_name
            formatted_operation_name = xform_name(operation_name)

            return self.perform_call(
                _client_make_api_call, 'client',
                service_name, formatted_operation_name,
                client, operation_name, api_params
            )

        return make_api_call_wrapper

    def _get_s3_call_wrapper(self, original_func):
        operation_name = original_func.__name__

        def s3_call_wrapper(*args, **kwargs):
            """Wraps __call__ of S3 client methods"""

            return self.perform_call(
                original_func, 'client', 's3', operation_name, *args, **kwargs
            )

        return s3_call_wrapper

    def perform_call(self, original_func, kind, service_name, operation_name,
                     *args, **kwargs):
        span = utils.start_child_span(
            operation_name='boto3:{}:{}:{}'.format(
                kind, service_name, operation_name
            ),
            parent=get_current_span()
        )

        span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
        span.set_tag(tags.COMPONENT, 'boto3')
        span.set_tag('boto3.service_name', service_name)

        with span, span_in_stack_context(span):
            try:
                response = original_func(*args, **kwargs)
            except ClientError as error:
                self.set_request_id_tag(span, error.response)
                raise
            else:
                if isinstance(response, dict):
                    self.set_request_id_tag(span, response)

        return response

    def _get_instrumented_executor_cls(self):
        class InstrumentedExecutor(_Executor):
            def submit(self, task, *args, **kwargs):
                return super(InstrumentedExecutor, self).submit(
                    keep_stack_context(task), *args, **kwargs
                )

        return InstrumentedExecutor


Boto3Patcher.configure_hook_module(globals())



# Copyright (c) 2015,2019 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import logging

from opentracing import tags


from .. import utils
from ._current_span import current_span_func
from ._patcher import Patcher

log = logging.getLogger(__name__)

try:
    from sqlalchemy.engine import Engine
    from sqlalchemy import event
except ImportError:
    pass


class SQLAlchemyPatcher(Patcher):
    applicable = 'event' in globals()

    def _install_patches(self):
        log.info('Instrumenting SQLAlchemy for tracing')
        event.listen(Engine, 'before_cursor_execute',
                     self.before_cursor_execute)
        event.listen(Engine, 'after_cursor_execute',
                     self.after_cursor_execute)

    def _reset_patches(self):
        event.remove(Engine, 'before_cursor_execute',
                     self.before_cursor_execute)
        event.remove(Engine, 'after_cursor_execute',
                     self.after_cursor_execute)

    @staticmethod
    def before_cursor_execute(conn, cursor, statement, parameters, context,
                              executemany):
        operation = 'SQL'
        statement = statement.strip()
        if statement:
            operation = '%s %s' % (operation,
                                   statement.split(' ', 1)[0].upper())
        span = utils.start_child_span(
            operation_name=operation, parent=current_span_func())
        span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
        span.set_tag(tags.COMPONENT, 'sqlalchemy')
        if conn.engine:
            span.set_tag(tags.DATABASE_TYPE, conn.engine.name)
            span.set_tag(tags.DATABASE_INSTANCE, repr(conn.engine.url))
        if statement:
            span.set_tag(tags.DATABASE_STATEMENT, statement)
        context.opentracing_span = span

    @staticmethod
    def after_cursor_execute(conn, cursor, statement, parameters, context,
                             executemany):
        if hasattr(context, 'opentracing_span') and context.opentracing_span:
            context.opentracing_span.finish()
            context.opentracing_span = None


SQLAlchemyPatcher.configure_hook_module(globals())



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from builtins import object
import threading

import opentracing
from opentracing.scope_managers.tornado import TornadoScopeManager
from opentracing.scope_managers.tornado import tracer_stack_context
from opentracing.scope_managers.tornado import ThreadSafeStackContext  # noqa


class RequestContext(object):
    """
    DEPRECATED, use either span_in_context() or span_in_stack_context()
    instead.

    RequestContext represents the context of a request being executed.

    Useful when a service needs to make downstream calls to other services
    and requires access to some aspects of the original request, such as
    tracing information.

    It is designed to hold a reference to the current OpenTracing Span,
    but the class can be extended to store more information.
    """

    __slots__ = ('span', )

    def __init__(self, span):
        self.span = span


class RequestContextManager(object):
    """
    DEPRECATED, use either span_in_context() or span_in_stack_context()
    instead.

    A context manager that saves RequestContext in thread-local state.

    Intended for use with ThreadSafeStackContext (a thread-safe
    replacement for Tornado's StackContext) or as context manager
    in a WSGI middleware.
    """

    _state = threading.local()
    _state.context = None

    @classmethod
    def current_context(cls):
        """Get the current request context.

        :rtype: opentracing_instrumentation.RequestContext
        :returns: The current request context, or None.
        """
        return getattr(cls._state, 'context', None)

    def __init__(self, context=None, span=None):
        # normally we want the context parameter, but for backwards
        # compatibility we make it optional and allow span as well
        if span:
            self._context = RequestContext(span=span)
        elif isinstance(context, opentracing.Span):
            self._context = RequestContext(span=context)
        else:
            self._context = context

    def __enter__(self):
        self._prev_context = self.__class__.current_context()
        self.__class__._state.context = self._context
        return self._context

    def __exit__(self, *_):
        self.__class__._state.context = self._prev_context
        self._prev_context = None
        return False


class _TracerEnteredStackContext(object):
    """
    An entered tracer_stack_context() object.

    Intended to have a ready-to-use context where
    Span objects can be activated before the context
    itself is returned to the user.
    """

    def __init__(self, context):
        self._context = context
        self._deactivation_cb = context.__enter__()

    def __enter__(self):
        return self._deactivation_cb

    def __exit__(self, type, value, traceback):
        return self._context.__exit__(type, value, traceback)


def get_current_span():
    """
    Access current request context and extract current Span from it.
    :return:
        Return current span associated with the current request context.
        If no request context is present in thread local, or the context
        has no span, return None.
    """
    # Check against the old, ScopeManager-less implementation,
    # for backwards compatibility.
    context = RequestContextManager.current_context()
    if context is not None:
        return context.span

    active = opentracing.tracer.scope_manager.active
    return active.span if active else None


def span_in_context(span):
    """
    Create a context manager that stores the given span in the thread-local
    request context. This function should only be used in single-threaded
    applications like Flask / uWSGI.

    ## Usage example in WSGI middleware:

    .. code-block:: python
        from opentracing_instrumentation.http_server import WSGIRequestWrapper
        from opentracing_instrumentation.http_server import before_request
        from opentracing_instrumentation import request_context

        def create_wsgi_tracing_middleware(other_wsgi):

            def wsgi_tracing_middleware(environ, start_response):
                request = WSGIRequestWrapper.from_wsgi_environ(environ)
                span = before_request(request=request, tracer=tracer)

                # Wrapper around the real start_response object to log
                # additional information to opentracing Span
                def start_response_wrapper(status, response_headers,
                                           exc_info=None):
                    if exc_info is not None:
                        span.log(event='exception', payload=exc_info)
                    span.finish()

                    return start_response(status, response_headers)

                with request_context.span_in_context(span):
                    return other_wsgi(environ, start_response_wrapper)

            return wsgi_tracing_middleware

    :param span: OpenTracing Span
    :return:
        Return context manager that wraps the request context.
    """

    # Return a no-op Scope if None was specified.
    if span is None:
        return opentracing.Scope(None, None)

    return opentracing.tracer.scope_manager.activate(span, False)


def span_in_stack_context(span):
    """
    Create Tornado's StackContext that stores the given span in the
    thread-local request context. This function is intended for use
    in Tornado applications based on IOLoop, although will work fine
    in single-threaded apps like Flask, albeit with more overhead.

    ## Usage example in Tornado application

    Suppose you have a method `handle_request(request)` in the http server.
    Instead of calling it directly, use a wrapper:

    .. code-block:: python

        from opentracing_instrumentation import request_context

        @tornado.gen.coroutine
        def handle_request_wrapper(request, actual_handler, *args, **kwargs)

            request_wrapper = TornadoRequestWrapper(request=request)
            span = http_server.before_request(request=request_wrapper)

            with request_context.span_in_stack_context(span):
                return actual_handler(*args, **kwargs)

    :param span:
    :return:
        Return StackContext that wraps the request context.
    """

    if not isinstance(opentracing.tracer.scope_manager, TornadoScopeManager):
        raise RuntimeError('scope_manager is not TornadoScopeManager')

    # Enter the newly created stack context so we have
    # storage available for Span activation.
    context = tracer_stack_context()
    entered_context = _TracerEnteredStackContext(context)

    if span is None:
        return entered_context

    opentracing.tracer.scope_manager.activate(span, False)
    assert opentracing.tracer.active_span is not None
    assert opentracing.tracer.active_span is span

    return entered_context



from __future__ import absolute_import

import opentracing
from opentracing.ext import tags

from ..request_context import get_current_span, span_in_context
from ._patcher import Patcher


try:
    from celery.app.task import Task
    from celery.signals import (
        before_task_publish, task_prerun, task_success, task_failure
    )
except ImportError:
    pass
else:
    _task_apply_async = Task.apply_async


def task_apply_async_wrapper(task, args=None, kwargs=None, **other_kwargs):
    operation_name = 'Celery:apply_async:{}'.format(task.name)
    span = opentracing.tracer.start_span(operation_name=operation_name,
                                         child_of=get_current_span())
    set_common_tags(span, task, tags.SPAN_KIND_RPC_CLIENT)

    with span_in_context(span), span:
        result = _task_apply_async(task, args, kwargs, **other_kwargs)
        span.set_tag('celery.task_id', result.task_id)
        return result


def set_common_tags(span, task, span_kind):
    span.set_tag(tags.SPAN_KIND, span_kind)
    span.set_tag(tags.COMPONENT, 'Celery')
    span.set_tag('celery.task_name', task.name)


def before_task_publish_handler(headers, **kwargs):
    headers['parent_span_context'] = span_context = {}
    opentracing.tracer.inject(span_context=get_current_span().context,
                              format=opentracing.Format.TEXT_MAP,
                              carrier=span_context)


def task_prerun_handler(task, task_id, **kwargs):
    request = task.request

    operation_name = 'Celery:run:{}'.format(task.name)
    child_of = None
    if request.delivery_info.get('is_eager'):
        child_of = get_current_span()
    else:
        if getattr(request, 'headers', None) is not None:
            # Celery 3.x
            parent_span_context = request.headers.get('parent_span_context')
        else:
            # Celery 4.x
            parent_span_context = getattr(request, 'parent_span_context', None)

        if parent_span_context:
            child_of = opentracing.tracer.extract(
                opentracing.Format.TEXT_MAP, parent_span_context
            )

    task.request.span = span = opentracing.tracer.start_span(
        operation_name=operation_name,
        child_of=child_of,
    )
    set_common_tags(span, task, tags.SPAN_KIND_RPC_SERVER)
    span.set_tag('celery.task_id', task_id)

    request.tracing_context = span_in_context(span)
    request.tracing_context.__enter__()


def finish_current_span(task, exc_type=None, exc_val=None, exc_tb=None):
    task.request.span.finish()
    task.request.tracing_context.__exit__(exc_type, exc_val, exc_tb)


def task_success_handler(sender, **kwargs):
    finish_current_span(task=sender)


def task_failure_handler(sender, exception, traceback, **kwargs):
    finish_current_span(
        task=sender,
        exc_type=type(exception),
        exc_val=exception,
        exc_tb=traceback,
    )


class CeleryPatcher(Patcher):
    applicable = '_task_apply_async' in globals()

    def _install_patches(self):
        Task.apply_async = task_apply_async_wrapper
        before_task_publish.connect(before_task_publish_handler)
        task_prerun.connect(task_prerun_handler)
        task_success.connect(task_success_handler)
        task_failure.connect(task_failure_handler)

    def _reset_patches(self):
        Task.apply_async = _task_apply_async
        before_task_publish.disconnect(before_task_publish_handler)
        task_prerun.disconnect(task_prerun_handler)
        task_success.disconnect(task_success_handler)
        task_failure.disconnect(task_failure_handler)


CeleryPatcher.configure_hook_module(globals())



# Copyright (c) 2017,2019 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from ._dbapi2 import ContextManagerConnectionWrapper as ConnectionWrapper
from ._dbapi2 import ConnectionFactory, CursorWrapper, NO_ARG
from ._singleton import singleton

# Try to save the original entry points
try:
    import psycopg2.extensions
    from psycopg2.sql import Composable
except ImportError:
    pass
else:
    _psycopg2_connect = psycopg2.connect
    _psycopg2_extensions_register_type = psycopg2.extensions.register_type
    _psycopg2_extensions_quote_ident = psycopg2.extensions.quote_ident


class Psycopg2CursorWrapper(CursorWrapper):
    """
    Psycopg2 accept not only string as sql statement, but instances of
    ``psycopg2.sql.Composable`` that should be represented as string before the
    executing.
    """
    def execute(self, sql, params=NO_ARG):
        if isinstance(sql, Composable):
            sql = sql.as_string(self)
        return super(Psycopg2CursorWrapper, self).execute(sql, params)

    def executemany(self, sql, seq_of_parameters):
        if isinstance(sql, Composable):
            sql = sql.as_string(self)
        return super(Psycopg2CursorWrapper, self).executemany(
            sql, seq_of_parameters
        )


@singleton
def install_patches():
    if 'psycopg2' not in globals():
        return

    # the following original methods checks a type of the conn_or_curs
    # and it doesn't accept wrappers
    def register_type(obj, conn_or_curs=None):
        if isinstance(conn_or_curs, (ConnectionWrapper, CursorWrapper)):
            conn_or_curs = conn_or_curs.__wrapped__
        _psycopg2_extensions_register_type(obj, conn_or_curs)

    def quote_ident(string, scope):
        if isinstance(scope, (ConnectionWrapper, CursorWrapper)):
            scope = scope.__wrapped__
        return _psycopg2_extensions_quote_ident(string, scope)

    psycopg2.extensions.register_type = register_type
    psycopg2.extensions.quote_ident = quote_ident

    factory = ConnectionFactory(connect_func=psycopg2.connect,
                                module_name='psycopg2',
                                conn_wrapper_ctor=ConnectionWrapper,
                                cursor_wrapper=Psycopg2CursorWrapper)
    setattr(psycopg2, 'connect', factory)
    if hasattr(psycopg2, 'Connect'):
        setattr(psycopg2, 'Connect', factory)




from .. import get_current_span

current_span_func = get_current_span


def set_current_span_func(span_extractor_func):
    """
    A convenience method to override the default method
      to extract parent span.
    It has to be called before install_patches
     :parent span_extractor_func : a function that returns parent span
    """
    global current_span_func
    current_span_func = span_extractor_func



class Patcher(object):

    def __init__(self):
        self.patches_installed = False

    @property
    def applicable(self):
        raise NotImplementedError

    def install_patches(self):
        if self.patches_installed:
            return

        if self.applicable:
            self._install_patches()
        self.patches_installed = True

    def reset_patches(self):
        if self.applicable:
            self._reset_patches()
        self.patches_installed = False

    def _install_patches(self):
        raise NotImplementedError

    def _reset_patches(self):
        raise NotImplementedError

    @classmethod
    def configure_hook_module(cls, context):
        def set_patcher(custom_patcher):
            context['patcher'] = custom_patcher

        def install_patches():
            context['patcher'].install_patches()

        def reset_patches():
            context['patcher'].reset_patches()

        context['patcher'] = cls()
        context['set_patcher'] = set_patcher
        context['install_patches'] = install_patches
        context['reset_patches'] = reset_patches



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import object
import logging
import urllib.parse
import opentracing
import six
from opentracing import Format
from opentracing.ext import tags
from opentracing_instrumentation import config


def before_request(request, tracer=None):
    """
    Attempts to extract a tracing span from incoming request.
    If no tracing context is passed in the headers, or the data
    cannot be parsed, a new root span is started.

    :param request: HTTP request with `.headers` property exposed
        that satisfies a regular dictionary interface
    :param tracer: optional tracer instance to use. If not specified
        the global opentracing.tracer will be used.
    :return: returns a new, already started span.
    """
    if tracer is None:
        tracer = opentracing.tracer

    # we need to prepare tags upfront, mainly because RPC_SERVER tag must be
    # set when starting the span, to support Zipkin's one-span-per-RPC model
    tags_dict = {
        tags.SPAN_KIND: tags.SPAN_KIND_RPC_SERVER,
        tags.HTTP_URL: request.full_url,
        tags.HTTP_METHOD: request.method,
    }

    remote_ip = request.remote_ip
    if remote_ip:
        tags_dict[tags.PEER_HOST_IPV4] = remote_ip

    caller_name = request.caller_name
    if caller_name:
        tags_dict[tags.PEER_SERVICE] = caller_name

    remote_port = request.remote_port
    if remote_port:
        tags_dict[tags.PEER_PORT] = remote_port

    operation = request.operation
    try:
        carrier = {}
        for key, value in six.iteritems(request.headers):
            carrier[key] = value
        parent_ctx = tracer.extract(
            format=Format.HTTP_HEADERS, carrier=carrier
        )
    except Exception as e:
        logging.exception('trace extract failed: %s' % e)
        parent_ctx = None

    span = tracer.start_span(
        operation_name=operation,
        child_of=parent_ctx,
        tags=tags_dict)

    return span


class AbstractRequestWrapper(object):
    """
    Exposes several properties used by the tracing methods.
    """

    @property
    def caller_name(self):
        for header in config.CONFIG.caller_name_headers:
            caller = self.headers.get(header.lower(), None)
            if caller is not None:
                return caller
        return None

    @property
    def full_url(self):
        raise NotImplementedError('full_url')

    @property
    def headers(self):
        raise NotImplementedError('headers')

    @property
    def method(self):
        raise NotImplementedError('method')

    @property
    def remote_ip(self):
        raise NotImplementedError('remote_ip')

    @property
    def remote_port(self):
        return None

    @property
    def server_port(self):
        return None

    @property
    def operation(self):
        return self.method


class TornadoRequestWrapper(AbstractRequestWrapper):
    """
    Wraps tornado.httputils.HTTPServerRequest and exposes several properties
    used by the tracing methods.
    """

    def __init__(self, request):
        self.request = request

    @property
    def full_url(self):
        return self.request.full_url()

    @property
    def headers(self):
        return self.request.headers

    @property
    def method(self):
        return self.request.method

    @property
    def remote_ip(self):
        return self.request.remote_ip


class WSGIRequestWrapper(AbstractRequestWrapper):
    """
    Wraps WSGI environment and exposes several properties
    used by the tracing methods.
    """

    def __init__(self, wsgi_environ, headers):
        self.wsgi_environ = wsgi_environ
        self._headers = headers

    @classmethod
    def from_wsgi_environ(cls, wsgi_environ):
        instance = cls(wsgi_environ=wsgi_environ,
                       headers=cls._parse_wsgi_headers(wsgi_environ))
        return instance

    @staticmethod
    def _parse_wsgi_headers(wsgi_environ):
        """
        HTTP headers are presented in WSGI environment with 'HTTP_' prefix.
        This method finds those headers, removes the prefix, converts
        underscores to dashes, and converts to lower case.

        :param wsgi_environ:
        :return: returns a dictionary of headers
        """
        prefix = 'HTTP_'
        p_len = len(prefix)
        # use .items() despite suspected memory pressure bc GC occasionally
        # collects wsgi_environ.iteritems() during iteration.
        headers = {
            key[p_len:].replace('_', '-').lower():
                val for (key, val) in wsgi_environ.items()
            if key.startswith(prefix)}
        return headers

    @property
    def full_url(self):
        """
        Taken from
        http://legacy.python.org/dev/peps/pep-3333/#url-reconstruction

        :return: Reconstructed URL from WSGI environment.
        """
        environ = self.wsgi_environ
        url = environ['wsgi.url_scheme'] + '://'

        if environ.get('HTTP_HOST'):
            url += environ['HTTP_HOST']
        else:
            url += environ['SERVER_NAME']

            if environ['wsgi.url_scheme'] == 'https':
                if environ['SERVER_PORT'] != '443':
                    url += ':' + environ['SERVER_PORT']
            else:
                if environ['SERVER_PORT'] != '80':
                    url += ':' + environ['SERVER_PORT']

        url += urllib.parse.quote(environ.get('SCRIPT_NAME', ''))
        url += urllib.parse.quote(environ.get('PATH_INFO', ''))
        if environ.get('QUERY_STRING'):
            url += '?' + environ['QUERY_STRING']
        return url

    @property
    def headers(self):
        return self._headers

    @property
    def method(self):
        return self.wsgi_environ.get('REQUEST_METHOD')

    @property
    def remote_ip(self):
        return self.wsgi_environ.get('REMOTE_ADDR', None)

    @property
    def remote_port(self):
        return self.wsgi_environ.get('REMOTE_PORT', None)

    @property
    def server_port(self):
        return self.wsgi_environ.get('SERVER_PORT', None)
